v.setCursorPosition(0,5);
v.type("$");
v.type("ok");

v.setCursorPosition(1,7);
v.type("<");
v.type("ok");

v.setCursorPosition(2,7);
v.type("{");
v.type("ok");

v.setCursorPosition(3,5);
v.type("$");
v.type("ok");
