v.setCursorPosition(0,4);
v.type("%");
v.type("ok;");

v.setCursorPosition(1,5);
v.type("%");
v.type("ok;");

v.setCursorPosition(2,5);
v.type("%");
v.type("ok;");

v.setCursorPosition(3,6);
v.type("%");
v.type("ok;");

v.setCursorPosition(4,6);
v.type("%");
v.type("ok;");

v.setCursorPosition(5,7);
v.type("%");
v.type("ok;");
