v.setCursorPosition(2,14);
v.enter();
v.type("public def foo");
v.enter();
v.type("puts \"in foo\"");
v.enter();
v.type("end");
v.enter();
v.enter();
v.type("protected def bar(x)");
v.enter();
v.type("puts \"in bar with #{x}\"");
v.enter();
v.type("end");
v.enter();
v.enter();
v.type("private def bat y");
v.enter();
v.type("puts \"in bat with #{y}\"");
v.enter();
v.type("end");
