v.setCursorPosition(2,2);
v.paste("const char * multilineindented = R\"sometext(\n             Here's some multiline text\n      whose indentation\n             must be preserved perfectly\n        like this\n)sometext\";");
